import os

import networkx as nx
import numpy as np
import copy
from numpy import pi
from typing import Union
import json

from qwak.Errors import StateOutOfBounds, UndefinedTimeList, EmptyProbDistList, MissingNodeInput
from qwak.State import State
from qwak.QuantumWalk import QuantumWalk
from qwak.ProbabilityDistribution import (
    ProbabilityDistribution,
)
from qwak.qwak import QWAK


class GraphicalQWAK:
    def __init__(
        self,
        staticN: int,
        dynamicN: int,
        staticGraph: nx.Graph,
        dynamicGraph: nx.Graph,
        staticStateList: list,
        dynamicStateList: list,
        staticTime: float,
        dynamicTimeList: list,
        qwakId: str = 'undefinedUser',
    ) -> None:
        """_summary_

        Parameters
        ----------
        staticN : int
            _description_
        dynamicN : int
            _description_
        staticGraph : nx.Graph
            _description_
        dynamicGraph : nx.Graph
            _description_
        staticStateList : list
            _description_
        dynamicStateList : list
            _description_
        staticTime : float
            _description_
        dynamicTimeList : list
            _description_
        """
        self._staticN = staticN
        self._dynamicN = dynamicN
        self._staticGraph = staticGraph
        self._dynamicGraph = dynamicGraph
        self._staticStateList = staticStateList
        self._dynamicStateList = dynamicStateList
        self._staticTime = staticTime
        self._dynamicTime = dynamicTimeList
        self._dynamicTimeList = np.linspace(
            self._dynamicTime[0], self._dynamicTime[1], int(
                self._dynamicTime[1]))
        self._staticQWAK = QWAK(self._staticGraph)
        # self._staticProbDist = self._staticQWAK.getProbDist()
        # self._dynamicQWAK = QWAK(self._dynamicGraph)
        # self._dynamicProbDistList = self._dynamicQWAK.getProbDistList()
        # self._dynamicAmpList = self._dynamicQWAK.getWalkList()
        self._qwakId = qwakId

    def to_json(self):
        """_summary_

        Returns
        -------
        str
            _description_
        """
        return json.dumps({
            'staticN': self._staticN,
            'dynamicN': self._dynamicN,
            'staticGraph': nx.node_link_data(self._staticGraph),
            'dynamicGraph': nx.node_link_data(self._dynamicGraph),
            'staticStateList': self._staticStateList,
            'dynamicStateList': self._dynamicStateList,
            'staticTime': self._staticTime,
            'dynamicTime': self._dynamicTime,
            'staticQWAK': json.loads(self._staticQWAK.to_json()),
            # 'dynamicQWAK': json.loads(self._dynamicQWAK.to_json()),
            # 'staticProbDist': self._staticProbDist.to_json(),
            # 'dynamicProbDistList':  list(map(lambda prob: prob.to_json(),self._dynamicProbDistList)),
            # 'dynamicAmpList': self._dynamicAmpList,
            'qwakId': self._qwakId
        })

    @classmethod
    def from_json(cls, json_var: Union[str, dict]):
        """_summary_

        Parameters
        ----------
        json_str : str
            _description_

        Returns
        -------
        Operator
            _description_
        """
        if isinstance(json_var, str):
            data = json.loads(json_var)
        elif isinstance(json_var, dict):
            data = json_var
        staticN = data['staticN']
        dynamicN = data['dynamicN']
        staticGraph = nx.node_link_graph(data['staticGraph'])
        dynamicGraph = nx.node_link_graph(data['dynamicGraph'])
        staticStateList = data['staticStateList']
        dynamicStateList = data['dynamicStateList']
        staticTime = data['staticTime']
        dynamicTime = data['dynamicTime']
        staticQWAK = QWAK.from_json(data['staticQWAK'])
        # dynamicQWAK = QWAK.from_json(data['dynamicQWAK'])
        # staticProbDist = ProbabilityDistribution.from_json(
        #     data['staticProbDist'])
        # dynamicProbDistList =  data['dynamicProbDistList']
        # dynamicAmpList =  data['dynamicAmpList']
        qwakId = data['qwakId']
        # print(staticQWAK.getProbDist().getDim())
        # print(f"Dict\t GQWAK {data['staticQWAK']['probDist']['state']['n']}")
        # print(f"Obj\t GQWAK {staticQWAK.getProbDist().getState().getDim()}")
        # print(staticN)
        # print(staticQWAK.getWalk().getFinalState())
        newGQwak = cls(
            staticN=staticN,
            dynamicN=dynamicN,
            staticGraph=staticGraph,
            dynamicGraph=dynamicGraph,
            staticStateList=staticStateList,
            dynamicStateList=dynamicStateList,
            staticTime=staticTime,
            dynamicTimeList=dynamicTime,
            qwakId=qwakId)
        # print(f"BEFORE SET {newGQwak.getStaticProbDist().getDim()}")
        # newGQwak.setDynamicTimeList(dynamicTimeList)
        newGQwak.setStaticQWAK(staticQWAK)
        # print(f"AFTER SET {newGQwak.getStaticProbDist().getDim()}")
        # newGQwak.setDynamicQWAK(dynamicQWAK)
        # newGQwak.setStaticProbDist(staticProbDist)
        # newGQwak.setDynamicProbDistList(dynamicProbDistList)
        # newGQwak.setDynamicWalkList(dynamicAmpList)

        # return ("nothing")
        return staticQWAK

    def getQwakId(self):
        return self._qwakId

    def setQwakId(self, newId):
        self._qwakId = newId

    def setStaticQWAK(self, newStaticQWAK):
        self._staticQWAK.setQWAK(newStaticQWAK)

    def getStaticQWAK(self):
        return self._staticQWAK

    def setDynamicQWAK(self, newDynamicQWAK):
        self._dynamicQWAK.setQWAK(newDynamicQWAK)

    def setDynamicTimeList(self, newTimeList):
        self._dynamicQWAK.setTimeList(newTimeList)
        self._dynamicTimeList = self._dynamicQWAK.getTimeList()

    def getStaticWalk(self):
        return self._staticQWAK.getWalk()

    def runWalk(self):
        """_summary_

        Returns
        -------
        _type_
            _description_
        """
        try:
            # self._staticQWAK.resetWalk()
            self._staticQWAK.runWalk(
                self._staticTime, self._staticStateList)
            self._staticProbDist = self._staticQWAK.getProbDist()
            qwProbVec = self._staticProbDist.getProbVec().tolist()
            return [False, qwProbVec]
        except StateOutOfBounds as err:
            return [True, str(err)]

    def runMultipleWalks(self):
        """_summary_

        Returns
        -------
        _type_
            _description_
        """
        try:
            self._dynamicQWAK.resetWalk()
            self._dynamicQWAK.runMultipleWalks(
                timeList=self._dynamicTimeList,
                initStateList=self._dynamicStateList[0])
            self._dynamicAmpList = self._dynamicQWAK.getWalkList()
            self._dynamicProbDistList = self._dynamicQWAK.getProbDistList()
            qwProbVecList = list(
                map(lambda probVec: probVec.tolist(), self._dynamicQWAK.getProbVecList()))
            return [False, qwProbVecList]
        except (StateOutOfBounds, UndefinedTimeList, EmptyProbDistList) as err:
            return [True, str(err)]

    def setStaticDim(self, newDim, graphStr, initStateList=None):
        """_summary_

        Parameters
        ----------
        newDim : _type_
            _description_
        graphStr : _type_
            _description_
        initStateList : _type_, optional
            _description_, by default None
        """
        self._staticN = newDim
        self._staticQWAK.setDim(
            self._staticN,
            graphStr=graphStr,
            initStateList=initStateList)
        self._staticGraph = self._staticQWAK.getGraph()

    def setDynamicDim(self, newDim, graphStr):
        """_summary_

        Parameters
        ----------
        newDim : _type_
            _description_
        graphStr : _type_
            _description_
        """
        self._dynamicN = newDim
        self._dynamicQWAK.setDim(self._dynamicN, graphStr)
        self._dynamicGraph = self._dynamicQWAK.getGraph()

    def getStaticDim(self):
        """_summary_

        Returns
        -------
        _type_
            _description_
        """
        return self._staticN

    def getDynamicDim(self):
        """_summary_

        Returns
        -------
        _type_
            _description_
        """
        return self._dynamicN

    def setStaticGraph(self, newGraphStr):
        """_summary_

        Parameters
        ----------
        newGraphStr : _type_
            _description_
        """
        self._staticGraph = eval(newGraphStr + f"({self._staticN})")
        self._staticN = len(self._staticGraph)
        self._staticQWAK.setGraph(self._staticGraph)

    def setDynamicGraph(self, newGraphStr):
        """_summary_

        Parameters
        ----------
        newGraphStr : _type_
            _description_
        """
        self._dynamicGraph = eval(newGraphStr + f"({self._dynamicN})")
        self._dynamicN = len(self._dynamicGraph)
        self._dynamicQWAK.setGraph(self._dynamicGraph)

    def setStaticCustomGraph(self, customAdjacency):
        """_summary_

        Parameters
        ----------
        customAdjacency : _type_
            _description_
        """
        self._staticGraph = nx.from_numpy_matrix(customAdjacency)
        self._staticQWAK.setGraph(self._staticGraph)
        self._staticN = len(self._staticGraph)
        self._staticStateList = [self._staticN // 2]
        self._staticQWAK.setDim(
            self._staticN,
            graph=self._staticGraph,
            initStateList=self._staticStateList)

    def setDynamicCustomGraph(self, customAdjacency):
        """_summary_

        Parameters
        ----------
        customAdjacency : _type_
            _description_
        """
        self._dynamicGraph = nx.from_numpy_matrix(customAdjacency)
        self._dynamicQWAK.setGraph(self._dynamicGraph)
        self._dynamicN = len(self._dynamicGraph)
        self._dynamicStateList = [[self._dynamicN // 2]]
        self._dynamicQWAK.setDim(
            self._dynamicN,
            graph=self._dynamicGraph,
            initStateList=self._dynamicStateList)

    def getStaticGraph(self):
        """_summary_

        Returns
        -------
        _type_
            _description_
        """
        return self._staticGraph

    def getDynamicGraph(self):
        """_summary_

        Returns
        -------
        _type_
            _description_
        """
        return self._dynamicGraph

    def getStaticGraphToJson(self):
        """_summary_

        Returns
        -------
        _type_
            _description_
        """
        return nx.cytoscape_data(self._staticGraph)

    def getDynamicGraphToJson(self):
        """_summary_

        Returns
        -------
        _type_
            _description_
        """
        return nx.cytoscape_data(self._dynamicGraph)

    def getStaticAdjacencyMatrix(self):
        """_summary_

        Returns
        -------
        _type_
            _description_
        """
        return self._staticQWAK.getAdjacencyMatrix()

    def getDynamicAdjacencyMatrix(self):
        """_summary_

        Returns
        -------
        _type_
            _description_
        """
        return self._dynamicQWAK.getAdjacencyMatrix()

    def setStaticTime(self, newTime):
        """_summary_

        Parameters
        ----------
        newTime : _type_
            _description_
        """
        self._staticTime = eval(newTime)
        self._staticQWAK.setTime(self._staticTime)

    def getStaticTime(self):
        """_summary_

        Returns
        -------
        _type_
            _description_
        """
        return self._staticTime

    def setDynamicTime(self, newTimeList):
        """_summary_

        Parameters
        ----------
        newTimeList : _type_
            _description_
        """
        parsedTime = list(map(float, newTimeList.split(",")))
        self._dynamicTimeList = np.linspace(
            parsedTime[0], parsedTime[1], int(
                parsedTime[1]))
        self._dynamicQWAK.setTimeList(self._dynamicTimeList)

    def getDynamicTime(self):
        """_summary_

        Returns
        -------
        _type_
            _description_
        """
        return self._dynamicTimeList

    def setDynamicInitStateList(self, newInitStateList):
        """_summary_

        Parameters
        ----------
        newInitStateList : _type_
            _description_
        """
        parsedInitState = newInitStateList.split(";")
        self._dynamicStateList = []
        for initState in parsedInitState:
            self._dynamicStateList.append(
                list(map(int, initState.split(","))))

    def setStaticInitState(self, initStateStr):
        """_summary_

        Parameters
        ----------
        initStateStr : _type_
            _description_
        """
        self._staticStateList = []
        self._staticStateList = list(map(int, initStateStr.split(",")))
        newState = State(self._staticQWAK.getDim())
        newState.buildState(self._staticStateList)
        self._staticQWAK.setInitState(newState)

    def getStaticInitState(self):
        """_summary_

        Returns
        -------
        _type_
            _description_
        """
        return self._staticStateList

    def getDynamicInitStateList(self):
        """_summary_

        Returns
        -------
        _type_
            _description_
        """
        return self._dynamicStateList

    def getStaticProbDist(self):
        """_summary_

        Returns
        -------
        _type_
            _description_
        """
        return self._staticProbDist

    def setStaticProbDist(self, newStaticProbDist):
        self._staticProbDist.setProbDist(newStaticProbDist)

    def getDynamicProbDistList(self):
        """_summary_

        Returns
        -------
        _type_
            _description_
        """
        return self._dynamicProbDistList

    def setDynamicProbDistList(self, newDynamicProbDistList):
        self._dynamicProbDistList = newDynamicProbDistList
        self._dynamicQWAK.setProbDistList(self._dynamicProbDistList)

    def getDynamicWalkList(self):
        return self._dynamicAmpList

    def setDynamicWalkList(self, newWalkList):
        self._dynamicAmpList = newWalkList
        self._dynamicQWAK.setWalkList(self._dynamicAmpList)

    def getStaticProbVec(self):
        """_summary_

        Returns
        -------
        _type_
            _description_
        """
        return self._staticProbDist.getProbVec()

    def getDynamicProbVecList(self):
        """_summary_

        Returns
        -------
        _type_
            _description_
        """
        return list(map(lambda probVec: probVec.tolist(),
                        self._dynamicQWAK.getProbVecList()))

    def getStaticMean(self):
        """_summary_

        Returns
        -------
        _type_
            _description_
        """
        return self._staticQWAK.getMean()

    def getDynamicMean(self):
        """_summary_

        Returns
        -------
        _type_
            _description_
        """
        meanList = []
        for probDist in self._dynamicProbDistList:
            meanList.append(probDist.mean())
        return meanList

    def getStaticSndMoment(self):
        """_summary_

        Returns
        -------
        _type_
            _description_
        """
        # TODO: This not used in the GUI.
        return self._staticQWAK.getSndMoment()

    def getDynamicSndMoment(self):
        """_summary_

        Returns
        -------
        _type_
            _description_
        """
        # TODO: This not used in the GUI.
        sndMomentList = []
        for probDist in self._dynamicProbDistList:
            sndMomentList.append(probDist.getSndMoment())
        return sndMomentList

    def getStaticStDev(self):
        """_summary_

        Returns
        -------
        _type_
            _description_
        """
        return self._staticQWAK.getStDev()

    def getDynamicStDev(self):
        """_summary_

        Returns
        -------
        _type_
            _description_
        """
        stDevList = []
        for probDist in self._dynamicProbDistList:
            stDevList.append(probDist.stDev())
        return stDevList

    def getStaticSurvivalProb(self, k0, k1):
        """_summary_

        Parameters
        ----------
        k0 : _type_
            _description_
        k1 : _type_
            _description_

        Returns
        -------
        _type_
            _description_
        """
        try:
            return [False, self._staticQWAK.getSurvivalProb(k0, k1)]
        except MissingNodeInput as err:
            return [True, str(err)]

    def getDynamicSurvivalProb(self, k0, k1):
        """_summary_

        Parameters
        ----------
        k0 : _type_
            _description_
        k1 : _type_
            _description_

        Returns
        -------
        _type_
            _description_
        """
        try:
            survProbList = []
            for probDist in self._dynamicProbDistList:
                survProbList.append(probDist.survivalProb(k0, k1))
            return [False, survProbList]
        except MissingNodeInput as err:
            return [True, str(err)]

    def getStaticInversePartRatio(self):
        """_summary_

        Returns
        -------
        _type_
            _description_
        """
        return self._staticQWAK.getInversePartRatio()

    def getDynamicInvPartRatio(self):
        """_summary_

        Returns
        -------
        _type_
            _description_
        """
        invPartRatioList = []
        for amps in self._dynamicAmpList:
            invPartRatioList.append(amps.invPartRatio())
        return invPartRatioList

    def checkPST(self, nodeA, nodeB):
        """_summary_

        Parameters
        ----------
        nodeA : _type_
            _description_
        nodeB : _type_
            _description_

        Returns
        -------
        _type_
            _description_
        """
        try:
            return [False, str(self._staticQWAK.checkPST(nodeA, nodeB))]
        except MissingNodeInput as err:
            return [True, str(err)]



from __future__ import annotations

import numpy as np
from scipy.linalg import inv
import json

from qwak.Errors import StateOutOfBounds, NonUnitaryState
from utils.jsonMethods import json_matrix_to_complex, complex_matrix_to_json


class State:
    """
    Class that represents the states that will be used in a quantum walk.
    States are represented by column vectors in quantum mechanics,
    therefore Numpy is used to generate ndarrays which contain these column vectors.
    """

    def __init__(self, n: int, nodeList: list = None,
                 customStateList: list = None) -> None:
        """Object is initialized with a mandatory user inputted dimension, an optional
        stateList parameter which will be used to create the amplitudes for each node in the state
        and an internal stateVec which will be a Numpy ndarray representing the column vector.

        Parameters
        ----------
        n : int
            Desired dimension of the state.
        nodeList : list, optional
            List containing what nodes will have uniform superposition in the state, by default None.
        customStateList : list, optional
            Custom amplitudes for the state, by default None.
        """
        self._n = n
        if nodeList is None:
            self._nodeList = []
        else:
            self._nodeList = nodeList
        if customStateList is None:
            self._customStateList = []
        else:
            self._customStateList = customStateList
        self._stateVec = np.zeros((self._n, 1), dtype=complex)

    def to_json(self):
        """In contrast, the to_json method is not marked with the @classmethod decorator because it is a method that is called on an instance of the Operator class. This means that it can access the attributes of the instance on which it is called, and it uses these attributes to generate the JSON string representation of the Operator instance. Since it requires access to the attributes of a specific Operator instance, it cannot be called on the Operator class itself.

        Returns
        -------
        _type_
            _description_
        """
        state_dict = {
            "n": self._n,
            "node_list": self._nodeList,
            "custom_state_list": self._customStateList,
            "state_vec": complex_matrix_to_json(
                self._stateVec.tolist()),
        }
        return json.dumps(state_dict)

    @classmethod
    def from_json(cls, json_var):
        """The from_json method is marked with the @classmethod decorator because it is a method that is called on the class itself, rather than on an instance of the class. This is necessary because it is used to create a new instance of the Operator class from a JSON string, and it does not require an instance of the Operator class to do so.

        Parameters
        ----------
        json_str : _type_
            _description_

        Returns
        -------
        _type_
            _description_
        """
        if isinstance(json_var, str):
            state_dict = json.loads(json_var)
        elif isinstance(json_var, dict):
            state_dict = json_var
        n = state_dict["n"]
        node_list = state_dict["node_list"]
        custom_state_list = state_dict["custom_state_list"]
        state_vec = np.array(
            json_matrix_to_complex(
                state_dict["state_vec"]),
            dtype=complex)
        state = cls(n, node_list, custom_state_list)
        state.setStateVec(state_vec)
        return state

    def buildState(
            self,
            nodeList: list = None,
            customStateList: list = None) -> None:
        """Builds state vector from state list, by creating a balanced superposition of all
        nodes in the nodeList.
        This will be changed in the future to make nodeList make more sense.

        Parameters
        ----------
        nodeList : list, optional
            List containing what nodes will have uniform superposition in the state, by default None.
        customStateList : list, optional
            Custom amplitudes for the state, by default None.
        """
        # TODO: We can probably find a better way to build this
        # function.\
        # self.resetState()
        if nodeList is not None:
            self.resetState()
            self._nodeList = nodeList
        if customStateList is not None:
            self.resetState()
            self._customStateList = customStateList
        if self._customStateList:
            self._checkUnitaryStateList(self._customStateList)
            for customState in self._customStateList:
                self._checkStateOutOfBounds(customState[0])
                self._stateVec[customState[0]] = customState[1]
        else:
            nodeAmp = np.sqrt(len(self._nodeList))
            for state in self._nodeList:
                self._checkStateOutOfBounds(state)
                self._stateVec[state] = 1 / nodeAmp

    def _checkStateOutOfBounds(self, node) -> None:
        """_summary_

        Parameters
        ----------
        node : _type_
            _description_

        Raises
        ------
        StateOutOfBounds
            _description_
        """
        if node >= self._n:
            raise StateOutOfBounds(
                f"State {node} is out of bounds for system of size {self._n} ([0-{self._n - 1}])."
            )

    def _checkUnitaryStateList(self, customStateList) -> None:
        """_summary_

        Parameters
        ----------
        customStateList : _type_
            _description_

        Raises
        ------
        NonUnitaryState
            _description_
        """
        unitaryState = 0
        for state in customStateList:
            unitaryState += np.abs(state[1]) ** 2
        unitaryState = round(unitaryState, 5)
        if unitaryState != float(1):
            raise NonUnitaryState(
                f"The sum of the square of the amplitudes is -- {unitaryState} -- instead of 1."
            )

    def herm(self):
        """_summary_

        Returns
        -------
        _type_
            _description_
        """
        return self._stateVec.H

    def inv(self):
        """_summary_

        Returns
        -------
        _type_
            _description_
        """
        return inv(self._stateVec)

    def resetState(self):
        """Resets the components of the State."""
        self._stateVec = np.zeros((self._n, 1), dtype=complex)

    def setDim(self, newDim: int, newNodeList: list = None) -> None:
        """Sets the current state dimension to a user defined one.

        Parameters
        ----------
        newDim : int
            New state dimension.
        """
        self._n = newDim
        self._stateVec = np.zeros((self._n, 1), dtype=complex)
        if newNodeList is not None:
            self._nodeList = newNodeList

    def getDim(self) -> int:
        """Gets the current state dimension.

        Returns
        -------
        int
            State dimension.
        """
        return self._n

    def setNodeList(self, newNodeList: list) -> None:
        """Sets current node list to a user inputted one.

        Parameters
        ----------
        newNodeList : list
            List containing the new nodes.
        """
        # TODO: Do we need this?
        self._nodeList = newNodeList

    def getNodeList(self) -> list:
        """Gets the current list of nodes.

        Returns
        -------
        list
            Current list of nodes.
        """
        return self._nodeList

    def setStateVec(self, newVec: np.ndarray) -> None:
        """Sets the column vector associated with the state to a user defined one.

        Parameters
        ----------
        newVec : np.ndarray
            New column vector for the state.
        """
        self._stateVec = newVec

    def getStateVec(self) -> np.ndarray:
        """Gets the column vector associated with the state.

        Returns
        -------
        np.ndarray
            Vector of the State.
        """
        return self._stateVec

    def setState(self, newState: State) -> None:
        """Sets all the parameters of the current state to user defined ones.

        Parameters
        ----------
        newState : State
            New state.
        """
        self._n = newState.getDim()
        self._nodeList = newState.getNodeList()
        self._stateVec = newState.getStateVec()

    def __mul__(self, other: np.ndarray) -> np.ndarray:
        """Left-side multiplication for the State class.

        Parameters
        ----------
        other : np.ndarray
            Another Numpy ndarray to multiply the state by.

        Returns
        -------
        np.ndarray
            Array of the multiplication
        """
        return self._stateVec * other

    def __rmul__(self, other: np.ndarray) -> np.ndarray:
        """Left-side multiplication for the State class.

        Parameters
        ----------
        other : np.ndarray
            Another Numpy ndarray to multiply the state by.

        Returns
        -------
        np.ndarray
            Array of the multiplication.
        """
        return other * self._stateVec

    def __matmul__(self, other):
        """_summary_

        Parameters
        ----------
        other : _type_
            _description_

        Returns
        -------
        _type_
            _description_
        """
        return self._stateVec @ other

    def __str__(self) -> str:
        """String representation of the State class.

        Returns
        -------
        str
            State string.
        """
        return f"{self._stateVec}"

    def __repr__(self) -> str:
        """Representation of the ProbabilityDistribution object.

        Returns
        -------
        str
            String of the ProbabilityDistribution object.
        """
        return f"N: {self._n}\n" \
               f"Node list: {self._nodeList}\n" \
               f"Custom Node list: {self._customStateList}\n" \
               f"State:\n\t{self._stateVec}"



from math import sqrt, ceil, pow

import numpy as np
from sympy import gcd, div, Float


def isStrCospec(A, a, b):
    """For a graph to have PST it needs to obey numerous rules and the first one is that the vertices with PST must be
    strongly cospectral. Two vertices a and b are strong cospectral if the characteristic polynomial of the matrix Ma, that has
    the colum and row a removed, is equal to the characteristic polynomial of the matrix Mb, that has the column and row b
    removed. This is easily checked using SymPy and it is what the function StrCospec does.

    Parameters
    ----------
    A : _type_
        _description_
    a : _type_
        _description_
    b : _type_
        _description_

    Returns
    -------
    _type_
        _description_
    """
    Ma = A.minor_submatrix(a, a)
    Mb = A.minor_submatrix(b, b)
    Mab = Ma.minor_submatrix(b - 1, b - 1)

    phi = A.charpoly()
    phia = Ma.charpoly()
    phib = Mb.charpoly()
    phiab = Mab.charpoly()

    if phia != phib or not isSimplePoles(phi, phiab):
        return False
    else:
        return True


def isSimplePoles(phi, phiab):
    """The second condition for strong cospectral is that the poles of phiab/phi must be simple. This is trickier to check and
    it starts by defning a polynomial g(x)=gcd(phi,phiab). Then, there will be no poles if f(x) = phi/g(x) is a polynomial
    without repeated roots. This occurs when gcd(f(x),f'(x)) is a constant, then we only need to check the coeeficients and
    see if all of them is zero except the last one. This is done by the function SimplePoles.

    Parameters
    ----------
    phi : _type_
        _description_
    phiab : _type_
        _description_

    Returns
    -------
    _type_
        _description_
    """
    gx = gcd(phi, phiab)
    fx, r = div(phi, gx)
    fx_der = fx.diff()
    great = gcd(fx, fx_der)
    coeffs = great.all_coeffs()
    del coeffs[len(coeffs) - 1]
    for coeff in coeffs:
        if coeff > 0:
            return False
    return True


def sieveEratosthenes(n):
    """One of the steps requires that we find square-free numbers, that is, numbers that have a prime decomposition with all
    primes being unique. Fist we use the Sieve of Eratosthenes, and the corresponding function, to find all primes from 1 to n.
    It works by choosing the first prime in the list, in this case 2, then squareing it and removing all its multiples from the
    list. The algorithm then choose the next prime and does the same procedure. At the end, we have a list with True at position
    i if i+1 is prime and False otherwise.

    Parameters
    ----------
    n : _type_
        _description_

    Returns
    -------
    _type_
        _description_
    """
    isPrime = np.full([n, 1], True)
    for i in range(2, ceil(sqrt(n)) + 1):
        if isPrime[i - 1]:
            for j in range(int(pow(i, 2)), n + 1, i):
                isPrime[j - 1] = False
    return isPrime


def getSquareFree(n):
    """The function SquareFree gives a list of square-free number from 1 to n in a similar way that we found all primes. The
    algorithm get the list of primes from 1 to n, then it start a loop of integers from 2 to n, when i is prime the algorithm
    associates False to all multiples of i^2. It returns a list with True (False) in the entry i if i-1 is (not) square free.

    Parameters
    ----------
    n : _type_
        _description_

    Returns
    -------
    _type_
        _description_
    """
    isSqrFree = np.full([n, 1], True)
    isPrime = sieveEratosthenes(n)
    sqrFreeList = []
    for i in range(2, n + 1):
        if isPrime[i - 1]:
            for j in range(int(pow(i, 2)), n + 1, int(pow(i, 2))):
                isSqrFree[j - 1] = False
    for i in range(n):
        if isSqrFree[i]:
            sqrFreeList.append(i + 1)
    return sqrFreeList


def getEigenSupp(a, eigenvec, eigenval):
    """EigenSupp is responsible for getting the eigenvalue support of the vertice a. The eigenvalue support of the vertice a is
    all the eigenvalues such that Er a != 0, where a is vector with 1 in the a-th entry and zero in all others and Er is the
    projection matrix of the eigenvalue r. It returns a list with all those eigenavlues. This is one of the conditions for PST,
    because we do not need to check for eigenvalues that are not in the eigenvalue support.

    Parameters
    ----------
    a : _type_
        _description_
    eigenvec : _type_
        _description_
    eigenval : _type_
        _description_

    Returns
    -------
    _type_
        _description_
    """
    supp = []
    for i in range(len(eigenval)):
        if eigenvec.col(i)[a] != 0:
            supp.append(eigenval[i])
    return supp


def checkRoots(A, a, eigenvec, eigenval):
    """CheckRoots is responsible for checking the second condition for PST which is that all eigenvalues is the eigenvalue support
    of a must be all integers or all quadratic integers with the format p+qr*Sqrt(delta)/2, with qr changing from eigenvalue to
    to eigenvalue. The frist step is to define h(x) = phi/gcd(phi,phia) that have all its roots in the eigenvalue support of a,
    then its degree k will be crucial.
    First, we check for integer roots in the interval [-n^4,n^4] that the eigenvalues should be. We check by putting the
    value in the loop, i, direct into h(i) and we see if it is equal to zero. If it is, then we check if i is in the
    eigenvalue support of a. With both conditions satisfied, it stores 1 to delta and sum one to intRoots.
    Then we check if all roots are quadratic integers p+qr*Sqrt(delta)/2. We know that p will be equal to the coefficient
    of the second bigest power of h(x). Then, all we need to do is loop through the values of delta in the list of square-free
    integers. Then, we loop qr until it is bigger than Sqrt(Tr(A²)) and check if is a root of h(x) and it is also in the
    eigenvalue support of a. In case it is true, we store the value of delta and sum one to quadRoots.
    Here we check if quadRoots or intRoots are bigger than k, which is the degree of our polynomial h(x). If none of them is
    then we know that PST is not possible and we return False.
    The time that PST occurs is just pi/g*Sqrt(delta) where g = gcd(theta0 - thetar), i.e. the gcd between all the differences
    of the eigenvalue theta0 (the biggest eigenvalue) and all others eigenvalues.

    Parameters
    ----------
    A : _type_
        _description_
    a : _type_
        _description_
    eigenvec : _type_
        _description_
    eigenval : _type_
        _description_

    Returns
    -------
    _type_
        _description_
    """
    supp = getEigenSupp(a, eigenvec, eigenval)
    Ma = Ma = A.minor_submatrix(a, a)
    phi = A.charpoly()
    phia = Ma.charpoly()
    h, r = div(phi, gcd(phi, phia))
    k = h.degree()
    intRoots = 0
    for i in range(-len(eigenval) ** 4, len(eigenval) ** 4 + 1):
        if h(i) == 0 and i in supp:
            delta = 1
            intRoots += 1
    quadRoots = 0
    sqrtFreeInt = getSquareFree(int((4 * A**2).trace()))
    p = h.all_coeffs()[1]
    deltaTmp = 0
    for deltaS in sqrtFreeInt:
        q = 0
        while (p + q * sqrt(deltaS) / 2) <= sqrt(int(((A**2).trace()))):
            if (
                h(p + q * sqrt(deltaS) / 2) == 0
                and Float(p + q * sqrt(deltaS) / 2, 3) in supp
            ):
                quadRoots += 1
                deltaTmp = deltaS
            q += 1

    if quadRoots > 0:
        delta = deltaTmp
    if quadRoots < k and intRoots < k:
        return False, 0, 0
    diffs = []
    for i in range(len(supp)):
        diffs.append((max(supp) - supp[i]) / np.sqrt(delta))

    g = 0
    for diff in diffs:
        g = np.gcd(g, diff)
    return True, g, delta


def swapNodes(nodeA, nodeB):
    """_summary_

    Parameters
    ----------
    nodeA : _type_
        _description_
    nodeB : _type_
        _description_

    Returns
    -------
    _type_
        _description_
    """
    nodeA = nodeA + nodeB
    nodeB = nodeA - nodeB
    nodeA = nodeA - nodeB
    return nodeA, nodeB


def getEigenVal(D):
    """_summary_

    Parameters
    ----------
    D : _type_
        _description_

    Returns
    -------
    _type_
        _description_
    """
    eigenVal = []
    for i in range(len(D.col(0))):
        temp = D.col(i)[i]
        if abs(temp) < 0.0000001:
            temp = 0
        eigenVal.append(temp)
    return eigenVal



from __future__ import annotations

import networkx as nx
import numpy as np
from qutip import Qobj, basis


class StochasticOperator(object):
    """
    Stochastic quantum walker on QuTip.
    Class containing an open quantum system described by a Lindblad equation obtained from the adjacency matrix.

    Theoretical model:
    Whitfield, J. D., Rodríguez-Rosario, C. A., & Aspuru-Guzik, A. (2010).
    Quantum stochastic walks: A generalization of classical random walks and quantum walks.
    Physical Review A, 81(2), 022323.

    @author: Lorenzo Buffoni
    """

    def __init__(
            self,
            graph,
            noiseParam=None,
            sinkNode=None,
            sinkRate=None) -> None:
        """_summary_

        Parameters
        ----------
        graph : _type_
            _description_
        noiseParam : _type_, optional
            _description_, by default None
        sinkNode : _type_, optional
            _description_, by default None
        sinkRate : _type_, optional
            _description_, by default None
        """
        self._graph = graph
        self.n = len(self._graph)
        self._adjacencyMatrix = (
            nx.adjacency_matrix(self._graph).todense().astype(complex)
        )
        self._buildLaplacian()
        if noiseParam is None:
            self._p = 0.0
        else:
            self._p = noiseParam
        if sinkRate is None:
            self._sinkRate = 1.0
        else:
            self._sinkRate = sinkRate
        self._sinkNode = sinkNode
        self._quantumHamiltonian = Qobj()
        self._classicalHamiltonian = []

    def buildStochasticOperator(
            self,
            noiseParam: float = None,
            sinkNode: int = None,
            sinkRate: float = None) -> None:
        """Creates the Hamiltonian and the Lindblad operators for the walker given an adjacency matrix
        and other parameters.

        Parameters
        ----------
        noiseParam : float, optional
            Parameter controlling the 'quantumness' of the system (0 is fully quantum, 1 is fully classical), by default None.
        sinkNode : int, optional
            _description_, by default None
        sinkRate : float, optional
            If a sink is present the trasfer rate from the sink_node to the sink , by default None.
        """
        if noiseParam is not None:
            self._p = noiseParam
        if sinkRate is not None:
            self._sinkRate = sinkRate
        if sinkNode is not None:
            self._sinkNode = sinkNode
        self._buildQuantumHamiltonian()
        self._buildClassicalHamiltonian()

    def _buildLaplacian(self) -> None:
        """_summary_"""
        degree = np.sum(self._adjacencyMatrix, axis=0).flat
        degree = list(map(lambda x: 1 / x if x > 0 else 0, degree))
        self._laplacian = np.multiply(self._adjacencyMatrix, degree)

    def _buildQuantumHamiltonian(self) -> None:
        """_summary_"""
        if self._sinkNode is not None:
            H = Qobj(
                (1 - self._p)
                * np.pad(self._adjacencyMatrix, [(0, 1), (0, 1)], "constant")
            )
        else:
            H = Qobj((1 - self._p) * self._adjacencyMatrix)
        self._quantumHamiltonian = H

    def _buildClassicalHamiltonian(self) -> None:
        """_summary_"""
        if self._sinkNode is not None:
            L = [
                np.sqrt(self._p * self._laplacian[i, j])
                * (basis(self.n + 1, i) * basis(self.n + 1, j).dag())
                for i in range(self.n)
                for j in range(self.n)
                if self._laplacian[i, j] > 0
            ]
            # transition matrix to the sink
            S = np.zeros([self.n + 1, self.n + 1])
            S[self.n, self._sinkNode] = np.sqrt(2 * self._sinkRate)
            L.append(Qobj(S))
        else:
            L = [
                np.sqrt(self._p * self._laplacian[i, j])
                * (basis(self.n, i) * basis(self.n, j).dag())
                for i in range(self.n)
                for j in range(self.n)
                if self._laplacian[i, j] > 0
            ]
        self._classicalHamiltonian = L

    def getClassicalHamiltonian(self) -> list:
        """_summary_

        Returns
        -------
        list
            _description_
        """
        return self._classicalHamiltonian

    def setClassicalHamiltonian(self, newClassicalHamiltonian) -> None:
        """_summary_

        Parameters
        ----------
        newClassicalHamiltonian : _type_
            _description_
        """
        self._classicalHamiltonian = newClassicalHamiltonian

    def getQuantumHamiltonian(self) -> Qobj:
        """_summary_

        Returns
        -------
        Qobj
            _description_
        """
        return self._quantumHamiltonian

    def setQuantumHamiltonian(self, newQuantumHamiltonian) -> None:
        """_summary_

        Parameters
        ----------
        newQuantumHamiltonian : _type_
            _description_
        """
        self._quantumHamiltonian = newQuantumHamiltonian

    def setSinkNode(self, newSinkNode) -> None:
        """_summary_

        Parameters
        ----------
        newSinkNode : _type_
            _description_
        """
        self._sinkNode = newSinkNode

    def getSinkNode(self) -> int:
        """_summary_

        Returns
        -------
        int
            _description_
        """
        return self._sinkNode

    def getLaplacian(self) -> np.ndarray:
        """_summary_

        Returns
        -------
        np.ndarray
            _description_
        """
        return self._laplacian



class Error(BaseException):
    """_summary_

    Parameters
    ----------
    BaseException : _type_
        _description_
    """

    def __init__(self, error_name, details):
        """_summary_

        Parameters
        ----------
        error_name : _type_
            _description_
        details : _type_
            _description_
        """
        self.error_name = error_name
        self.details = details

    def as_string(self):
        """_summary_

        Returns
        -------
        _type_
            _description_
        """
        return f"{self.error_name}: {self.details}\n"


class StateOutOfBounds(Error):
    """_summary_

    Parameters
    ----------
    Error : _type_
        _description_
    """

    def __init__(self, details):
        """_summary_

        Parameters
        ----------
        details : _type_
            _description_
        """
        super().__init__("Condition out of bounds: ", details)


class NonUnitaryState(Error):
    """_summary_

    Parameters
    ----------
    Error : _type_
        _description_
    """

    def __init__(self, details):
        """_summary_

        Parameters
        ----------
        details : _type_
            _description_
        """
        super().__init__("State is not unitary: ", details)


class UndefinedTimeList(Error):
    """_summary_

    Parameters
    ----------
    Error : _type_
        _description_
    """

    def __init__(self, details):
        """_summary_

        Parameters
        ----------
        details : _type_
            _description_
        """
        super().__init__("Time interval for multiple walks is undefined: ", details)


class EmptyProbDistList(Error):
    """_summary_

    Parameters
    ----------
    Error : _type_
        _description_
    """

    def __init__(self, details):
        """_summary_

        Parameters
        ----------
        details : _type_
            _description_
        """
        super().__init__("Probability distribution list is empty: ", details)


class MissingNodeInput(Error):
    """_summary_

    Parameters
    ----------
    Error : _type_
        _description_
    """

    def __init__(self, details):
        """_summary_

        Parameters
        ----------
        details : _type_
            _description_
        """
        super().__init__("Input nodes required: ", details)


class MissingGraphInput(Error):
    """_summary_

    Parameters
    ----------
    Error : _type_
        _description_
    """

    def __init__(self, details):
        """_summary_

        Parameters
        ----------
        details : _type_
            _description_
        """
        super().__init__("Graph required: ", details)



import networkx as nx
import numpy as np
import copy
import json
from qwak.Errors import (
    StateOutOfBounds,
    NonUnitaryState,
    UndefinedTimeList,
    EmptyProbDistList,
    MissingNodeInput,
    MissingGraphInput,
)
from qwak.State import State
from qwak.Operator import Operator
from qwak.QuantumWalk import QuantumWalk
from qwak.ProbabilityDistribution import (
    ProbabilityDistribution,
)


class QWAK:
    """Data access class that combines all three components required to
    perform a continuous-time quantum walk, given by the multiplication of
    an operator (represented by the Operator class) by an initial state
    (State class).  This multiplication is achieved in the
    StaticQuantumwalk class, which returns a final state (State Class)
    representing the amplitudes of each state associated with a graph node.
    These amplitudes can then be transformed to probability distributions
    (ProbabilityDistribution class) suitable for plotting with matplotlib,
    or your package of choice.
    """

    def __init__(
        self,
        graph: nx.Graph,
        time: float = None,
        timeList: list = None,
        initStateList: list = None,
        customStateList: list = None,
        laplacian: bool = False,
        markedSearch: list = None,
        qwakId: str = 'userUndef',
    ) -> None:
        """Default values for the initial state, time and transition rate are a
        column vector full of 0s, 0 and 1, respectively. Methods runWalk or
        buildWalk must then be used to generate the results of the quantum
        walk.

        Parameters
        ----------
        graph : nx.Graph
            NetworkX graph where the walk takes place. Also used
            for defining the dimensions of the quantum walk.
        time : float
            _ Needs completion _
        timeList : list
            _ Needs completion _
        initStateList : list[int], optional
            List with chosen initial states for uniform superposition, by default None
        customStateList : list[(int,complex)], optional
            Custom init state, by default None.
        laplacian : bool, optional
            Allows the user to choose whether to use the
            Laplacian or simple adjacency matrix, by default False.
        markedSearch : list, optional
            List with marked elements for search, by default None.
        """
        self._graph = graph
        self._n = len(self._graph)
        if timeList is not None:
            self._timeList = np.array(timeList)
        else:
            self._timeList = np.zeros(self._n)
        self._qwakId = qwakId
        self._operator = Operator(
            self._graph,
            time=time,
            laplacian=laplacian,
            markedSearch=markedSearch)
        self._initState = State(
            self._n,
            nodeList=initStateList,
            customStateList=customStateList)
        self._quantumWalk = QuantumWalk(self._initState, self._operator)

        self._probDist = ProbabilityDistribution(
            self._quantumWalk.getFinalState())

        self._probDistList = []

    def to_json(self, isDynamic=False) -> str:
        """_summary_

        Returns
        -------
        str
            _description_
        """
        if isDynamic:
            qwakJson = json.dumps({
                'dim': self._n,
                'graph': nx.node_link_data(self._graph),
                'timeList': self._timeList.tolist(),
                'initState': json.loads(self._initState.to_json()),
                'operator': json.loads(self._operator.to_json()),
                'quantumWalk': json.loads(self._quantumWalk.to_json()),
                # 'probDist': json.loads(self._probDist.to_json()),
                'probDistList': [probDist.to_json() for probDist in self._probDistList],
                # 'walkList': self._walkList,
                'qwakId': self._qwakId
            })
        else:
            qwakJson = json.dumps({
                'dim': self._n,
                'graph': nx.node_link_data(self._graph),
                # 'timeList': self._timeList.tolist(),
                'initState': json.loads(self._initState.to_json()),
                'operator': json.loads(self._operator.to_json()),
                'quantumWalk': json.loads(self._quantumWalk.to_json()),
                'probDist': json.loads(self._probDist.to_json()),
                # 'probDistList': list(map(lambda prob: json.loads(prob.to_json()),self._probDistList)),
                # 'walkList': self._walkList,
                'qwakId': self._qwakId
            })
        return qwakJson

    @classmethod
    def from_json(cls, json_var: str, isDynamic=False):
        """_summary_

        Parameters
        ----------
        json_str : str
            _description_

        Returns
        -------
        Operator
            _description_
        """
        if isinstance(json_var, str):
            data = json.loads(json_var)
        elif isinstance(json_var, dict):
            data = json_var
        qwakId = data['qwakId']
        graph = nx.node_link_graph(data['graph'])

        initState = State.from_json(data['initState'])
        operator = Operator.from_json(data['operator'])
        quantumWalk = QuantumWalk.from_json(data['quantumWalk'])
        if isDynamic:
            timeList = data['timeList']
            newQwak = cls(graph=graph, timeList=timeList, qwakId=qwakId)
            probDistList = [ProbabilityDistribution.from_json(
                probDist) for probDist in data['probDistList']]
            newQwak.setProbDistList(probDistList)
        else:
            probDist = ProbabilityDistribution.from_json(
                data['probDist'])
            newQwak = cls(graph=graph, qwakId=qwakId)
            newQwak.setProbDist(probDist)
        newQwak.setInitState(initState)
        newQwak.setOperator(operator)
        newQwak.setWalk(quantumWalk)
        return newQwak

    def setQWAK(self, newQWAK):
        self.setGraph(newQWAK.getGraph())
        self.setDim(newQWAK.getDim(), graph=self._graph)
        self.setInitState(newQWAK.getInitState())
        self.setOperator(newQWAK.getOperator())
        self.setWalk(newQWAK.getWalk())
        self.setProbDist(newQWAK.getProbDist())

    def runWalk(
            self,
            time: float = None,
            initStateList: list = None,
            customStateList: list = None) -> None:
        """Builds class' attributes, runs the walk and calculates the amplitudes
        and probability distributions with the given parameters. These can be
        accessed with their respective get methods.

        Parameters
        ----------
        time : float, optional
            Time for which to calculate the quantum walk, by default 0.
        initStateList : list[int], optional
            List with chosen initial states for uniform superposition, by default None.
        customStateList : list[(int,complex)], optional
            Custom init state, by default None.

        Raises
        ------
        stOBErr
            _description_
        nUErr
            _description_
        """
        try:
            self._initState.buildState(
                nodeList=initStateList, customStateList=customStateList
            )
        except StateOutOfBounds as stOBErr:
            raise stOBErr
        except NonUnitaryState as nUErr:
            raise nUErr

        self._operator.buildDiagonalOperator(time=time)
        self._quantumWalk.buildWalk(self._initState, self._operator)
        self._probDist.buildProbDist(self._quantumWalk.getFinalState())

    def runMultipleWalks(
            self,
            timeList: list = None,
            initStateList: list = None,
            customStateList: list = None) -> None:
        """_summary_

        Parameters
        ----------
        timeList : list, optional
            _description_, by default None
        initStateList : list, optional
            _description_, by default None
        customStateList : list, optional
            _description_, by default None

        Raises
        ------
        UndefinedTimeList
            _description_
        """
        self._probDistList = []
        if timeList is not None:
            self._timeList = timeList
        elif self._timeList is None:
            raise UndefinedTimeList(f"TimeList is {self._timeList}.")
        for time in self._timeList:
            self.runWalk(
                time=time,
                initStateList=initStateList,
                customStateList=customStateList)
            self._probDistList.append(copy.deepcopy(self.getProbDist()))

    def resetWalk(self) -> None:
        """Resets the components of a walk."""
        self._initState.resetState()
        self._operator.resetOperator()
        self._quantumWalk.resetWalk()
        self._probDist.resetProbDist()
        self._probDistList = []
        self._walkList = []

    def setDim(
            self,
            newDim: int,
            graphStr: str = None,
            graph: nx.Graph = None,
            initStateList: list = None) -> None:
        """Sets the current walk dimensions to a user defined one.
        Also takes a graph string to be
        evaluated and executed as a NetworkX graph generator.

        Parameters
        ----------
        newDim : int
            New dimension for the quantum walk.
        graphStr : str
            Graph string to generate the graph with the new dimension.
        initStateList : list[int], optional
            Init state list with new dimension.
        """
        # TODO: Docs need update after rework.
        self._n = newDim
        if graphStr is not None:
            self._graph = eval(f"{graphStr}({self._n})")
            # We need a reassignment here since certain graphs have
            # different length than the number inputted.
            self._n = len(self._graph)
        elif graph is not None:
            self._graph = graph
            self._n = len(self._graph)
        else:
            raise MissingGraphInput(
                f"You tried to set QWAK dim without providing a graph with updated dimensions: {self._graph}")

        self._initState.setDim(newDim, newNodeList=initStateList)
        self._operator.setDim(newDim, self._graph)
        self._quantumWalk.setDim(newDim)
        self._probDist.setDim(newDim)

    def getDim(self) -> int:
        """Gets the current graph dimension.
        Returns
        -------
        int
            Dimension of graph.
        """
        return self._n

    def setGraph(self, newGraph: nx.Graph, initStateList=None) -> None:
        """Sets the current graph to a user defined one.
        Also recalculates the current operator and walk dimension.

        Parameters
        ----------
        newGraph : nx.Graph
            New NetworkX graph.
        """
        self._graph = newGraph
        self._n = len(self._graph)
        self.setDim(self._n, graph=self._graph)

    def getGraph(self) -> nx.Graph:
        """Gets the current graph.

        Returns
        -------
        nx.Graph
            Current graph.
        """
        return self._graph

    def setCustomGraph(self, customAdjMatrix):
        self._graph = nx.from_numpy_matrix(customAdjMatrix)
        self.setGraph(newGraph=self._graph)
        self._initStateList = [self._n // 2]
        self.setDim(
            self._n,
            graph=self._graph,
            initStateList=self._initStateList)

    def setInitState(self, newInitState: State) -> None:
        """Sets the current initial state to a user defined one.

        Parameters
        ----------
        newInitState : State
            New initial state.
        """
        self._initState.setState(newInitState)
        self._initStateList = self._initState.getNodeList()

    def getInitState(self) -> State:
        """Gets the initial state.

        Returns
        -------
        State
            Initial State.
        """
        return self._initState

    def setTime(self, newTime: float) -> None:
        """Sets the current walk time to a user defined one.

        Parameters
        ----------
        newTime : float
            New time.
        """
        self._operator.setTime(newTime)

    def setTimeList(self, newTimeList: list) -> None:
        """_summary_

        Parameters
        ----------
        newTime : list
            _description_
        """
        timeList = np.linspace(
            newTimeList[0], newTimeList[1], int(
                newTimeList[1]))
        self._timeList = timeList

    def getTime(self) -> float:
        """Gets the current walk time.

        Returns
        -------
        float
           Current value of time.
        """
        return self._operator.getTime()

    def getTimeList(self) -> float:
        """Gets the current walk time.

        Returns
        -------
        float
           Current value of time.
        """
        return self._timeList

    def setAdjacencyMatrix(
            self, newAdjMatrix: np.ndarray, initStateList: list = None
    ) -> None:
        """_summary_

        Parameters
        ----------
        newAdjMatrix : np.ndarray
            _description_
        initStateList : list, optional
            _description_, by default None
        """
        self._n = len(self._operator.getAdjacencyMatrix())
        self._operator.setAdjacencyMatrix(newAdjMatrix)
        self._initState = State(self._n, initStateList)
        self._quantumWalk = QuantumWalk(self._initState, self._operator)
        self._probDist = ProbabilityDistribution(
            self._quantumWalk.getFinalState())

    def getAdjacencyMatrix(self) -> np.ndarray:
        """_summary_

        Returns:
            np.ndarray: _description_
        """
        return self._operator.getAdjacencyMatrix()

    def setOperator(self, newOperator: Operator) -> None:
        """Sets the current walk operator a user defined one.

        Parameters
        ----------
        newOperator : Operator
            New operator object.
        """
        self._operator.setOperator(newOperator)

    def getOperator(self) -> Operator:
        """Gets the current walk operator.

        Returns
        -------
        Operator
            Current operator object.
        """
        return self._operator

    def setWalk(self, newWalk: State) -> None:
        """Sets current walk amplitudes to a user defined state.
        This might not be needed and removed in the future.

        Parameters
        ----------
        newWalk : State
            New walk amplitudes.
        """
        self._quantumWalk.setWalk(newWalk)

    def getWalk(self) -> QuantumWalk:
        """Gets current QuantumWalk object

        Returns
        -------
        QuantumWalk
            Current state amplitudes.
        """
        return self._quantumWalk

    def getWalkList(self) -> list:
        """_summary_

        Returns
        -------
        list
            _description_
        """
        return self._walkList

    def setWalkList(self, newWalkList):
        self._walkList = newWalkList

    def getFinalState(self) -> State:
        """Gets current QuantumWalk State.

        Returns
        -------
        State
            State of the QuantumWalk.
        """
        return self._quantumWalk.getFinalState()

    def getAmpVec(self) -> np.ndarray:
        """Gets the array of the QuantumWalk state.

        Returns
        -------
        np.ndarray
            Array of the QuantumWalk state.
        """
        return self._quantumWalk.getAmpVec()

    def getAmpVecList(self) -> list:
        """_summary_

        Returns
        -------
        list
            _description_
        """
        return list(map(lambda x: x.getAmpVec(), self._walkList))

    def setProbDist(self, newProbDist: ProbabilityDistribution) -> None:
        """Sets current walk probability distribution to a user defined one.
        This might not be needed and removed in the future.

        Parameters
        ----------
        newProbDist : ProbabilityDistribution
            New probability distribution.
        """
        self._probDist.setProbDist(newProbDist)

    def getProbDist(self) -> ProbabilityDistribution:
        """Gets the current probability distribution.

        Returns
        -------
        ProbabilityDistribution
            ProbabilityDistribution object.
        """
        return self._probDist

    def getProbDistList(self) -> list:
        """_summary_

        Returns
        -------
        _type_
            _description_
        """
        if not self._probDistList:
            raise EmptyProbDistList(
                f"Prob. dist. list is {self._probDistList}. Perhaps you didnt run multiple walks?")
        return self._probDistList

    def setProbDistList(self, newProbDistList):
        self._probDistList = newProbDistList

    def getProbVec(self) -> np.ndarray:
        """Gets the current probability distribution vector.

        Returns
        -------
        np.ndarray
            Probability Distribution vector.
        """
        return self._probDist.getProbVec()

    def getProbVecList(self) -> list:
        """_summary_

        Returns
        -------
        _type_
            _description_
        """
        return [probDist.getProbVec()
                for probDist in self._probDistList]

    def searchNodeAmplitude(self, searchNode: int) -> complex:
        """User inputted node for search

        Parameters
        ----------
        searchNode : int
            User inputted node for the search.

        Returns
        -------
        complex
            Amplitude associated with the search node.
        """
        return self._quantumWalk.searchNodeAmplitude(searchNode)

    def searchNodeProbability(self, searchNode: int) -> float:
        """Searches and gets the probability associated with a given node.

        Parameters
        ----------
        searchNode : int
            User inputted node for the search.

        Returns
        -------
        float
            Probability associated with the search node.
        """
        return self._probDist.searchNodeProbability(searchNode)

    def getMean(self, resultRounding=None):
        """_summary_

        Returns
        -------
        _type_
            _description_
        """
        return self._probDist.moment(1) if (
            resultRounding is None) \
            else round(self._probDist.moment(1), resultRounding)

    def getMeanList(self, resultRounding=None):

        return [
            probDist.moment(1) for probDist in self._probDistList] if (
            resultRounding is None) \
            else [
            round(
                probDist.moment(1),
                resultRounding) for probDist in self._probDistList]

    def getSndMoment(self, resultRounding=None):
        """_summary_

        Returns
        -------
        _type_
            _description_
        """
        return self._probDist.moment(2) if (resultRounding is None) \
            else round(
            self._probDist.moment(2), resultRounding)

    def getStDev(self, resultRounding=None):
        """_summary_

        Returns
        -------
        _type_
            _description_
        """
        return self._probDist.stDev() if (resultRounding is None) \
            else round(
            self._probDist.stDev(),
            resultRounding)

    def getStDevList(self, resultRounding=None):
        return [
            probDist.stDev() for probDist in self._probDistList] if (
            resultRounding is None) \
            else [
            round(
                probDist.stDev(),
                resultRounding) for probDist in self._probDistList]

    def getInversePartRatio(self):
        return self._probDist.invPartRatio()

    def getInversePartRatioList(self, resultRounding=0):
        stDevList = [round(probDist.invPartRatio(), resultRounding)
                     for probDist in self._probDistList]
        return stDevList

    def getSurvivalProb(self, k0, k1):
        """_summary_

        Parameters
        ----------
        k0 : _type_
            _description_
        k1 : _type_
            _description_

        Returns
        -------
        _type_
            _description_
        """
        try:
            return self._probDist.survivalProb(k0, k1)
        except MissingNodeInput as err:
            raise err

    def getSurvivalProbList(self, k0, k1, resultRounding=0):
        try:
            return [
                probDist.survivalProb(
                    k0,
                    k1) for probDist in self._probDistList] if (resultRounding is None) \
            else [
                round(
                    probDist.survivalProb(
                        k0,
                        k1),
                    resultRounding) for probDist in self._probDistList]
        except MissingNodeInput as err:
            raise err

    def checkPST(self, nodeA, nodeB):
        """_summary_

        Parameters
        ----------
        nodeA : _type_
            _description_
        nodeB : _type_
            _description_

        Returns
        -------
        _type_
            _description_
        """
        try:
            return self._operator.checkPST(nodeA, nodeB)
        except MissingNodeInput as err:
            raise err

    def getTransportEfficiency(self):
        """_summary_

        Returns
        -------
        _type_
            _description_
        """
        return self._quantumWalk.transportEfficiency()



import numpy as np
from qutip import Qobj


class StochasticProbabilityDistribution(object):
    """_summary_

    Parameters
    ----------
    object : _type_
        _description_
    """

    # TODO: Figure out why we need the object.
    def __init__(self, state: Qobj) -> None:
        """_summary_

        Parameters
        ----------
        state : Qobj
            Initial state which will be the basis of the time dependant evolution.
        """
        self._finalState = state.getFinalState()
        self._n = state.getDim()
        self._probVec = np.zeros((self._n, 1))

    def buildProbDist(self, state=None) -> None:
        """_summary_

        Parameters
        ----------
        state : _type_, optional
            _description_, by default None
        """
        if state is not None:
            self._finalState = state.getFinalState()
        self._probVec = np.diagonal(self._finalState)

    def getProbVec(self) -> np.ndarray:
        """_summary_

        Returns
        -------
        np.ndarray
            _description_
        """
        return self._probVec.flatten()

    def setProbVec(self, newFinalState) -> None:
        """_summary_

        Parameters
        ----------
        newFinalState : _type_
            _description_
        """
        self._finalState = newFinalState



import warnings

import numpy as np
from qwak.State import State
from qwak.Errors import MissingNodeInput
import json
from utils.jsonMethods import json_matrix_to_complex, complex_matrix_to_json

warnings.filterwarnings("ignore")


class ProbabilityDistribution:
    """
    Class containing the vector containing the probabilities associated with the
    final state of a continuous-time quantum walk.

    """

    def __init__(self, state: State) -> None:
        """The dimension of the probability vector will then be loaded from
        the state inputted by the user.
        The vector containing the probabilities will be initialized full of zeros
        with the dimension obtained from the state.

        Parameters
        ----------
        state : State
            State to be converted into a probability.
        """
        self._state = state
        self._stateVec = self._state.getStateVec()
        self._n = state.getDim()
        self._probVec = np.zeros(self._n, dtype=float)

    def to_json(self):
        return json.dumps({
            'state': json.loads(self._state.to_json()),
            'dim': self._n,
            'prob_vec': self._probVec.tolist()
        })

    @classmethod
    def from_json(cls, json_var):
        if isinstance(json_var, str):
            json_dict = json.loads(json_var)
        elif isinstance(json_var, dict):
            json_dict = json_var
        state = State.from_json(json_dict['state'])
        prob_vec = np.array(json_dict['prob_vec'])
        probDist = cls(state)
        probDist.setProbVec(prob_vec)
        return probDist

    def resetProbDist(self) -> None:
        """Resets the ProbabilityDistribution object."""
        # TODO: Rethink state attribute
        self._stateVec = np.zeros((self._n, 1), dtype=complex)
        self._probVec = np.zeros(self._n, dtype=float)

    def buildProbDist(self, state: State = None) -> None:
        """Builds the probability vector by multiplying the user inputted
        amplitude state by its conjugate.

        Parameters
        ----------
        state : State, optional
            _description_, by default None
        """
        if state is not None:
            self._n = state.getDim()
            self._state.setState(state)
            self._stateVec = self._state.getStateVec()
        self._probVec = np.array(
            [(state[0] * np.conj(state[0])).real for state in self._stateVec])

    def setProbDist(self, newProbDist) -> None:
        """_summary_

        Parameters
        ----------
        newProbDist : _type_
            _description_
        """
        self._n = newProbDist.getDim()
        self._state.setState(newProbDist.getState())
        self._stateVec = newProbDist.getStateVec()
        self._probVec = newProbDist.getProbVec()

    def getStateVec(self) -> State:
        """_summary_

        Returns
        -------
        State
            _description_
        """
        return self._stateVec

    def getState(self):
        return self._state

    def setState(self, newState):
        self._state.setState(newState)

    def setDim(self, newDim):
        self._n = newDim
        self._probVec = np.zeros((self._n, 1), dtype=float)

    def getDim(self) -> int:
        """_summary_

        Returns
        -------
        int
            _description_
        """
        return self._n

    def setProbVec(self, newProbVec: np.ndarray) -> None:
        """Sets the current probability vector to a user inputted one.

        Parameters
        ----------
        newProbVec : np.ndarray
            New probability vector for the distribution.
        """
        self._probVec = newProbVec

    def getProbVec(self) -> np.ndarray:
        """Gets the probability vector associated with a distribution.

        Returns
        -------
        np.ndarray
            Returns the array of the ProbabilityDistribution object.
        """
        return self._probVec

    def searchNodeProbability(self, searchNode: int) -> float:
        """Searches and gets the probability associated with a given node.

        Parameters
        ----------
        searchNode : int
            User inputted node for the search.

        Returns
        -------
        float
            Probability of the searched node.
        """
        return self._probVec.item(searchNode)

    def mean(self) -> float:
        """Gets the mean of the current probability distribution.

        Returns
        -------
        float
            Mean of the ProbabilityDistribution object.
        """
        # TODO: This function has been replaced by moment(1).
        pos = np.arange(0, self._n)
        m = 0
        for x in range(self._n):
            m += pos[x] * self._probVec[x]
        return float(m)

    def moment(self, k) -> float:
        """_summary_

        Parameters
        ----------
        k : _type_
            _description_

        Returns
        -------
        float
            _description_
        """
        pos = np.arange(0, self._n)
        m = 0
        for x in range(self._n):
            m += (pos[x] ** k) * self._probVec[x]
        return float(m)

    def invPartRatio(self):
        invPartRatio = 0
        for prob in self._probVec:
            invPartRatio += np.absolute(prob.item(0))**2
        invPartRatio = 1 / invPartRatio
        return invPartRatio

    def stDev(self) -> float:
        """_summary_

        Returns
        -------
        float
            _description_
        """
        stDev = self.moment(2) - self.moment(1) ** 2
        if stDev <= 0:
            return 0
        return np.sqrt(stDev)

    def survivalProb(self, k0, k1) -> float:
        """_summary_

        Parameters
        ----------
        k0 : _type_
            _description_
        k1 : _type_
            _description_

        Returns
        -------
        float
            _description_
        """
        survProb = 0
        try:
            k0 = int(k0)
            k1 = int(k1)
            if k0 == k1:
                return self._probVec[int(k0)][0]
            else:
                for i in range(int(k0), int(k1) + 1):
                    survProb += self._probVec[i]
            return survProb
        except ValueError:
            raise MissingNodeInput(
                f"A node number is missing: k0 = {k0}; k1={k1}")

    def __str__(self) -> str:
        """String representation of the ProbabilityDistribution object.

        Returns
        -------
        str
            String of the ProbabilityDistribution object.
        """
        return f"{self._probVec}"

    def __repr__(self) -> str:
        """Representation of the ProbabilityDistribution object.

        Returns
        -------
        str
            String of the ProbabilityDistribution object.
        """
        return f"N: {self._n}\n" \
               f"State:\n\t{self._stateVec}\n" \
               f"ProbDist:\n\t{self._probVec}"



from __future__ import annotations

import networkx as nx
import sympy as sp
from scipy.linalg import inv
from sympy.abc import pi
import numpy as np
from utils.jsonMethods import json_matrix_to_complex, complex_matrix_to_json
import json

from qwak.Errors import MissingNodeInput
from utils.PerfectStateTransfer import isStrCospec, checkRoots, swapNodes, getEigenVal


class Operator:
    """
    Class that represents the operators that will be used in a quantum walk.
    States are represented by matrices in quantum mechanics,
    therefore Numpy is used to generate ndarrays which contain these matrices.
    """

    def __init__(
            self,
            graph: nx.Graph,
            time: float = None,
            laplacian: bool = False,
            markedSearch: list = None,
    ) -> None:
        """This object is initialized with a user inputted graph, which is then used to
        generate the dimension of the operator and the adjacency matrix, which is
        the central structure required to perform walks on regular graphs. Note that this
        version of the software only supports regular undirected graphs, which will hopefully
        be generalized in the future.

        The eigenvalues and eigenvectors of the adjacency matrix are also calculated during
        initialization, which are then used to calculate the diagonal operator through spectral
        decomposition. This was the chosen method since it is computationally cheaper than calculating
        the matrix exponent directly.

        Parameters
        ----------
        graph : nx.Graph
            Graph where the walk will be performed.
        laplacian : bool, optional
            Allows the user to choose whether to use the Laplacian or simple adjacency matrix, by default False.
        markedSearch : list, optional
            List with marked elements for search, by default None.
        """
        if time is not None:
            self._time = time
        else:
            self._time = 0
        if laplacian is not None:
            self._laplacian = laplacian
        else:
            self._laplacian = False
        if markedSearch is not None:
            self._markedSearch = markedSearch
        else:
            self._markedSearch = None
        self._graph = graph
        self._buildAdjacency(self._laplacian, self._markedSearch)
        self._n = len(graph)
        self._operator = np.zeros((self._n, self._n), dtype=complex)
        self._isHermitian = np.allclose(
            self._adjacencyMatrix,
            self._adjacencyMatrix.conjugate().transpose())
        self._buildEigenValues(self._isHermitian)

    def to_json(self) -> str:
        """_summary_

        Returns
        -------
        str
            _description_
        """
        return json.dumps({
            'graph': nx.node_link_data(self._graph),
            'time': self._time,
            'laplacian': self._laplacian,
            'markedSearch': self._markedSearch,
            'adjacencyMatrix': complex_matrix_to_json(self._adjacencyMatrix),
            'eigenvalues': self._eigenvalues.tolist(),
            'eigenvectors': complex_matrix_to_json(self._eigenvectors),
            'operator': complex_matrix_to_json(self._operator),
        })

    @classmethod
    def from_json(cls, json_var: str) -> Operator:
        """_summary_

        Parameters
        ----------
        json_str : str
            _description_

        Returns
        -------
        Operator
            _description_
        """
        if isinstance(json_var, str):
            data = json.loads(json_var)
        elif isinstance(json_var, dict):
            data = json_var
        graph = nx.node_link_graph(data['graph'])
        time = data['time']
        laplacian = data['laplacian']
        markedSearch = data['markedSearch']
        adjacencyMatrix = np.array(
            json_matrix_to_complex(
                data['adjacencyMatrix']))
        eigenvalues = np.array(data['eigenvalues'])

        eigenvectors = np.array(
            json_matrix_to_complex(
                data['eigenvectors']))
        operator = np.array(json_matrix_to_complex(data['operator']))

        newOp = cls(graph, time, laplacian, markedSearch)
        newOp._setAdjacencyMatrixOnly(adjacencyMatrix)
        newOp._setEigenValues(eigenvalues)
        newOp._setEigenVectors(eigenvectors)
        newOp._setOperatorVec(operator)
        return newOp

    def buildDiagonalOperator(self, time: float = None) -> None:
        """Builds operator matrix from optional time and transition rate parameters, defined by user.
        The first step is to calculate the diagonal matrix that takes in time, transition rate and
        eigenvalues and convert it to a list of the diagonal entries. The entries are then multiplied
        by the eigenvectors, and the last step is to perform matrix multiplication with the complex
        conjugate of the eigenvectors.

        Parameters
        ----------
        time : float, optional
            Time for which to calculate the operator, by default 0.
        """
        if time is not None:
            self._time = time
        diag = np.diag(
            np.exp(-1j * self._eigenvalues * self._time)).diagonal()
        self._operator = np.multiply(self._eigenvectors, diag)
        if self._isHermitian:
            self._operator = np.matmul(
                self._operator,
                self._eigenvectors.conjugate().transpose())
        else:
            self._operator = np.matmul(
                self._operator, inv(
                    self._eigenvectors))

    def _buildAdjacency(self, laplacian, markedSearch):
        """_summary_

        Parameters
        ----------
        laplacian : _type_
            _description_
        markedSearch : _type_
            _description_
        """
        if laplacian:
            self._adjacencyMatrix = np.asarray(
                nx.laplacian_matrix(
                    self._graph).todense().astype(complex))
        else:
            self._adjacencyMatrix = nx.to_numpy_array(
                self._graph, dtype=complex)
        if markedSearch is not None:
            for marked in markedSearch:
                self._adjacencyMatrix[marked[0], marked[0]] += marked[1]

    def _buildEigenValues(self, isHermitian):
        """_summary_

        Parameters
        ----------
        isHermitian : bool
            _description_
        """
        if isHermitian:
            self._eigenvalues, self._eigenvectors = np.linalg.eigh(
                self._adjacencyMatrix
            )
        else:
            self._eigenvalues, self._eigenvectors = np.linalg.eig(
                self._adjacencyMatrix)

    def getEigenValues(self):
        """_summary_

        Returns
        -------
        _type_
            _description_
        """
        return self._eigenvalues

    def _setEigenValues(self, eigenValues):
        """_summary_

        Parameters
        ----------
        eigenValues : _type_
            _description_
        """
        self._eigenvalues = eigenValues

    def _setEigenVectors(self, eigenVectors):
        """_summary_

        Parameters
        ----------
        eigenVectors : _type_
            _description_
        """
        self._eigenvectors = eigenVectors

    def getEigenVectors(self):
        """_summary_

        Returns
        -------
        _type_
            _description_
        """
        return self._eigenvectors

    def resetOperator(self):
        """Resets Operator object."""
        self._operator = np.zeros((self._n, self._n), dtype=complex)

    def setDim(self, newDim: int, graph) -> None:
        """Sets the current Operator objects dimension to a user defined one.

        Parameters
        ----------
        newDim : int
            New dimension for the Operator object.
        """
        self._n = newDim
        self._operator = np.zeros((self._n, self._n), dtype=complex)
        self._graph = graph
        self._adjacencyMatrix = (
            nx.adjacency_matrix(self._graph).todense().astype(complex)
        )
        self.setAdjacencyMatrix(self._adjacencyMatrix)

    def getDim(self) -> int:
        """Gets the current graph dimension.

        Returns
        -------
        int
            Dimension of Operator object.
        """
        return self._n

    def setTime(self, newTime: float) -> None:
        """Sets the current operator time to a user defined one.

        Parameters
        ----------
        newTime : float
            New operator time.
        """
        self._time = newTime

    def getTime(self) -> float:
        """Gets the current operator time.

        Returns
        -------
        float
            Current time of Operator object.
        """
        return self._time

    def setAdjacencyMatrix(self, adjacencyMatrix: np.ndarray) -> None:
        """Sets the adjacency matrix of the operator to a user defined one.
        Might make more sense to not give the user control over this parameter, and make
        them instead change the graph entirely.

        Parameters
        ----------
        adjacencyMatrix : np.ndarray
            New adjacency matrix.
        """
        self._adjacencyMatrix = adjacencyMatrix.astype(complex)
        self._n = len(self._adjacencyMatrix)
        self.resetOperator()
        self._buildEigenValues(self._isHermitian)

    def _setAdjacencyMatrixOnly(
            self, adjacencyMatrix: np.ndarray) -> None:
        """Sets the adjacency matrix of the operator to a user defined one.
        Might make more sense to not give the user control over this parameter, and make
        them instead change the graph entirely.

        Parameters
        ----------
        adjacencyMatrix : np.ndarray
            New adjacency matrix.
        """
        self._adjacencyMatrix = adjacencyMatrix.astype(complex)
        self._n = len(self._adjacencyMatrix)
        self.resetOperator()

    def getAdjacencyMatrix(self) -> np.ndarray:
        """Gets the current adjacency matrix of the Operator.

        Returns
        -------
        np.ndarray
            Adjacency matrix of the Operator.
        """
        return self._adjacencyMatrix

    def _setOperatorVec(self, newOperator: np.ndarray) -> None:
        """Sets all the parameters of the current operator to user defined ones.

        Parameters
        ----------
        newOperator : Operator
            New user inputted Operator.
        """
        self._operator = newOperator

    def setOperator(self, newOperator: Operator) -> None:
        """Sets all the parameters of the current operator to user defined ones.

        Parameters
        ----------
        newOperator : Operator
            New user inputted Operator.
        """
        self._n = newOperator.getDim()
        self._time = newOperator.getTime()
        self._operator = newOperator.getOperator()

    def getOperator(self) -> np.matrix:
        """Gets the numpy matrix associated with the current operator.

        Returns
        -------
        np.matrix
            Current Operator object.
        """
        return self._operator

    def checkPST(self, nodeA, nodeB):
        """Algorithm to check PST based on the article https://arxiv.org/abs/1606.02264 authored by Rodrigo Chaves.
        Checks if all the conditions are true and return the **VALUE** if the graph
        has PST and False otherwise.

        Parameters
        ----------
        nodeA : _type_
            Input node.
        nodeB : _type_
            Output node.

        Returns
        -------
        Float/Bool
            Either returns the time value of PST or False.
        """
        # TODO: Check if numpy is faster for eigenvecs and vals.
        try:
            nodeA = int(nodeA)
            nodeB = int(nodeB)
            if nodeA > nodeB:
                nodeA, nodeB = swapNodes(nodeA, nodeB)

            symAdj = sp.Matrix(self._adjacencyMatrix.tolist())
            # Isto já foi calculado.
            eigenVec, D = symAdj.diagonalize()
            eigenVal = getEigenVal(D)
            isCospec = isStrCospec(symAdj, nodeA, nodeB)
            chRoots, g, delta = checkRoots(
                symAdj, nodeA, eigenVec, eigenVal)
            if isCospec and chRoots:
                result = pi / (g * np.sqrt(delta))
            else:
                result = -1
            return result
        except ValueError:
            raise MissingNodeInput(
                f"A node number is missing: nodeA = {nodeA}; nodeB = {nodeB}")

    def __mul__(self, other: np.ndarray) -> np.ndarray:
        """Left-side multiplication for the Operator class.

        Parameters
        ----------
        other : np.ndarray
            Another Numpy ndarray to multiply the operator by.

        Returns
        -------
        np.ndarray
            Result of the right-side multiplication.
        """
        return self._operator * other

    def __rmul__(self, other: np.ndarray) -> np.ndarray:
        """Right-side multiplication for the Operator class.

        Parameters
        ----------
        other : np.ndarray
            Another Numpy ndarray to multiply the operator by.

        Returns
        -------
        np.ndarray
            Result of the left-side multiplication.
        """
        return other * self._operator

    def __str__(self) -> str:
        """String representation of the State class.

        Returns
        -------
        str
            String representation of the Operator object.
        """
        return f"{self._operator}"

    def __repr__(self) -> str:
        """Representation of the ProbabilityDistribution object.

        Returns
        -------
        str
            String of the ProbabilityDistribution object.
        """
        return f"N: {self._n}\n" \
               f"Time: {self._time}\n" \
               f"Graph: {nx.to_dict_of_dicts(self._graph)}\n" \
               f"Operator:\n\t{self._operator}"

    # def transportEfficiency(self,initState):
    #     """
    #     Under Construction.
    #     @param initState:
    #     @return:
    #     """
    #     ef = 0
    #     print(f"init: {initState}")
    #     print(f"Eigenvectors {self._eigenvectors}")
    #     for i in range(len(self._eigenvectors)):
    #         eigenVec = np.transpose(self._eigenvectors[:,i]).conjugate()
    #         ef += np.absolute(np.matmul(eigenVec,initState))**2
    #         print(f"eigenVec: {eigenVec}\t\t eigenVec norm: {np.linalg.norm(eigenVec)}\t\tef : {ef}\n")
    #     return ef



import networkx as nx
import numpy as np
from qutip import Options

from qwak.Errors import StateOutOfBounds, NonUnitaryState
from qwak.State import State
from qwak.StochasticOperator import StochasticOperator
from qwak.StochasticProbabilityDistribution import StochasticProbabilityDistribution
from qwak.StochasticQuantumWalk import StochasticQuantumWalk


class StochasticQWAK:
    """_summary_"""

    def __init__(
        self,
        graph: nx.Graph,
        initStateList=None,
        customStateList=None,
        noiseParam=None,
        sinkNode=None,
        sinkRate=None,
    ) -> None:
        """_summary_

        Parameters
        ----------
        graph : nx.Graph
            _description_
        initStateList : _type_, optional
            _description_, by default None
        customStateList : _type_, optional
            _description_, by default None
        noiseParam : _type_, optional
            _description_, by default None
        sinkNode : _type_, optional
            _description_, by default None
        sinkRate : _type_, optional
            _description_, by default None
        """
        self._graph = graph
        self._n = len(self._graph)
        self._operator = StochasticOperator(
            self._graph,
            noiseParam=noiseParam,
            sinkNode=sinkNode,
            sinkRate=sinkRate,
        )
        self._initState = State(
            self._n,
            nodeList=initStateList,
            customStateList=customStateList)
        self._quantumWalk = StochasticQuantumWalk(
            self._initState, self._operator)
        self._probDist = StochasticProbabilityDistribution(
            self._quantumWalk)

    def runWalk(
        self,
        time: float = 0,
        initStateList: list = None,
        customStateList=None,
        noiseParam=None,
        sinkNode=None,
        sinkRate=None,
        observables=[],
        opts=Options(store_states=True, store_final_state=True),
    ) -> None:
        """_summary_

        Parameters
        ----------
        time : float, optional
            _description_, by default 0
        initStateList : list, optional
            _description_, by default None
        customStateList : _type_, optional
            _description_, by default None
        noiseParam : _type_, optional
            _description_, by default None
        sinkNode : _type_, optional
            _description_, by default None
        sinkRate : _type_, optional
            _description_, by default None
        observables : list, optional
            _description_, by default []
        opts : _type_, optional
            _description_, by default Options(store_states=True, store_final_state=True)

        Raises
        ------
        stOBErr
            _description_
        nUErr
            _description_
        """
        # TODO: Move the constructors to the constructor method.
        try:
            self._initState.buildState(
                nodeList=initStateList, customStateList=customStateList
            )
        except StateOutOfBounds as stOBErr:
            raise stOBErr
        except NonUnitaryState as nUErr:
            raise nUErr
        self._operator.buildStochasticOperator(
            noiseParam=noiseParam, sinkNode=sinkNode, sinkRate=sinkRate
        )
        self._quantumWalk = StochasticQuantumWalk(
            self._initState, self._operator)
        self._quantumWalk.buildWalk(time, observables, opts)
        self._probDist = StochasticProbabilityDistribution(
            self._quantumWalk)
        self._probDist.buildProbDist()

    def setProbDist(self, newProbDist: object) -> None:
        """_summary_

        Parameters
        ----------
        newProbDist : object
            _description_
        """
        self._probDist = newProbDist

    def getProbDist(self) -> StochasticProbabilityDistribution:
        """_summary_

        Returns
        -------
        ProbabilityDistribution
            _description_
        """
        return self._probDist

    def getProbVec(self) -> np.ndarray:
        """_summary_

        Returns
        -------
        ProbabilityDistribution
            _description_
        """
        return self._probDist.getProbVec()

    def getQuantumHamiltonian(self):
        """_summary_

        Returns
        -------
        _type_
            _description_
        """
        return self._operator.getQuantumHamiltonian()

    def getClassicalHamiltonian(self):
        """_summary_

        Returns
        -------
        _type_
            _description_
        """
        return self._operator.getClassicalHamiltonian()

    def getLaplacian(self):
        """_summary_

        Returns
        -------
        _type_
            _description_
        """
        return self._operator.getLaplacian()



def complex_to_json(z):
    if isinstance(z, complex):
        return {'real': z.real, 'imag': z.imag}
    else:
        type_name = z.__class__.__name__
        raise TypeError(f"Object of type '{type_name}' is not JSON serializable")

def complex_matrix_to_json(complex_matrix):
    jsonList = []
    jsonListAux = []
    if not isinstance(complex_matrix,list):
        complex_matrix = complex_matrix.tolist()
    for cmplist in complex_matrix:
        for cmp in cmplist:
            jsonListAux.append(complex_to_json(cmp))
        jsonList.append(jsonListAux)
        jsonListAux = []
    return jsonList

def json_to_complex(dct):
    if 'real' in dct and 'imag' in dct:
        return complex(dct['real'], dct['imag'])
    return dct

def json_matrix_to_complex(json_matrix):
    listComplex = []
    listComplexAux = []
    for dct in json_matrix:
        for d in dct:
            listComplexAux.append(json_to_complex(d))
        listComplex.append(listComplexAux)
        listComplexAux = []
    return listComplex



from __future__ import annotations

import warnings
import numpy as np
import json
from utils.jsonMethods import json_matrix_to_complex, complex_matrix_to_json

from qwak.Operator import Operator
from qwak.State import State


warnings.filterwarnings("ignore")


class QuantumWalk:
    """Class that represents the final state containing the amplitudes of a
    continuous-time quantum walk.
    """

    def __init__(self, state: State, operator: Operator) -> None:
        """This object is initialized with a user inputted initial state and
        operator.
        The dimension of the quantum walk will then be loaded from the initial
        state.
        The final state will contain the amplitudes of the time evolution of
        the initial state, as a function of the operator. This variable is initialized
        as an instance of State class.

        Parameters
        ----------
        state : State
            Initial state which will be the basis of the time dependant evolution.
        operator : Operator
            Operator which will evolve the initial state.
        """
        # TODO: This class has mandatory init conditions. This is not in
        # line with the rest of the classes.
        self._n = state.getDim()
        self._initState = state
        self._operator = operator
        self._finalState = State(self._n)

    def to_json(self) -> str:
        """Serializes the QuantumWalk object to JSON format.

        Returns
        -------
        str
            JSON string representation of the QuantumWalk object.
        """
        return json.dumps({
            "n": self._n,
            "initState": json.loads(self._initState.to_json()),
            "operator": json.loads(self._operator.to_json()),
            "finalState": json.loads(self._finalState.to_json())
        })

    @classmethod
    def from_json(cls, json_var: str) -> QuantumWalk:
        """Deserializes a JSON string to a QuantumWalk object.

        Parameters
        ----------
        json_str : str
            JSON string representation of the QuantumWalk object.

        Returns
        -------
        QuantumWalk
            Deserialized QuantumWalk object.
        """
        if isinstance(json_var, str):
            data = json.loads(json_var)
        elif isinstance(json_var, dict):
            data = json_var
        initState = State.from_json(data["initState"])
        operator = Operator.from_json(data["operator"])
        finalState = State.from_json(data["finalState"])
        walk = cls(initState, operator)
        walk.setFinalState(finalState)
        return walk

    def buildWalk(self, initState: State = None,
                  operator: Operator = None) -> None:
        """Builds the final state of the quantum walk by setting it to the matrix
        multiplication of the operator by the initial state.

        Parameters
        ----------
        initState : State, optional
            Initial state which will be the basis of the time dependant evolution, by default None.
        operator : Operator, optional
            Operator which will evolve the initial state, by default None.
        """
        # TODO: Name of init state variable not in line with class init.
        if initState is not None:
            self._initState = initState
        if operator is not None:
            self._operator = operator
        self._finalState.setStateVec(
            np.matmul(
                self._operator.getOperator(),
                self._initState.getStateVec()))

    def resetWalk(self):
        """Resets the components of the QuantumWalk object."""
        self._operator.resetOperator()
        self._initState.resetState()
        self._finalState.resetState()

    def setInitState(self, newInitState: State) -> None:
        """Sets the initial state of the quantum walk to a new user inputted one.

        Parameters
        ----------
        newInitState : State
            New initial state for the quantum walk.
        """
        self._initState.setState(newInitState)

    def getInitState(self) -> State:
        """Gets the initial state of the quantum walk.

        Returns
        -------
        State
            Initial state of the quantum walk.
        """
        return self._initState

    def setDim(self, newDim: int) -> None:
        """Sets the current quantum walk dimension to a user defined one.

        Parameters
        ----------
        newDim : int
            New QuantumWalk dimension.
        """
        self._n = newDim
        self._finalState.setDim(self._n)

    def getDim(self) -> int:
        """Gets the current state dimension.

        Returns
        -------
        int
            QuantumWalk dimension.
        """
        return self._n

    def setOperator(self, newOperator: Operator) -> None:
        """Sets the current operator to a user defined one.

        Parameters
        ----------
        newOperator : Operator
            New quantum walk operator.
        """
        self._operator.setOperator(newOperator)

    def getOperator(self) -> Operator:
        """Gets the current operator.

        Returns
        -------
        Operator
            Current QuantumWalk Operator object.
        """
        return self._operator

    def setWalk(self, newWalk: QuantumWalk) -> None:
        """Sets all the parameters of the current quantum walk to user defined ones.

        Parameters
        ----------
        newWalk : QuantumWalk
            New quantum walk.
        """
        self._initState.setState(newWalk.getInitState())
        self._operator.setOperator(newWalk.getOperator())
        self._finalState.setState(newWalk.getFinalState())

    def getFinalState(self) -> State:
        """Gets the final state of the QuantumWalk.

        Returns
        -------
        State
            Final state of the QuantumWalk.
        """
        return self._finalState

    def setFinalState(self, newFinalState):
        """Sets the final state of the QuantumWalk.

        Parameters
        -------
        finalState
            Final state of the QuantumWalk.
        """
        self._finalState.setState(newFinalState)

    def getAmpVec(self) -> np.ndarray:
        """Gets the vector of the final state amplitudes of the  QuantumWalk.

        Returns
        -------
        np.ndarray
            Vector of the final state.
        """
        # TODO: This function name is not in line with getFinalState.
        return self._finalState.getStateVec()

    def searchNodeAmplitude(self, searchNode: int) -> complex:
        """Searches and gets the amplitude associated with a given node.

        Parameters
        ----------
        searchNode : int
            User inputted node for the search.

        Returns
        -------
        complex
            Amplitude of the search node.
        """
        return self._finalState.getStateVec().item(searchNode)

    def invPartRatio(self) -> float:
        """_summary_

        Returns
        -------
        float
            _description_
        """
        amplitudes = 0
        for amp in self._finalState.getStateVec():
            amplitudes += np.absolute(amp.item(0)) ** 4
        amplitudes = amplitudes
        return 1 / amplitudes

    def transportEfficiency(self) -> float:
        """_summary_

        Returns
        -------
        float
            _description_
        """
        return 1 - np.trace(self._finalState @ self._finalState.herm())

    def __str__(self) -> str:
        """String representation of the StaticQuantumwalk class.

        Returns
        -------
        str
            QuantumWalk string.
        """
        return f"{self._finalState.getStateVec()}"

    def __repr__(self) -> str:
        """Representation of the ProbabilityDistribution object.

        Returns
        -------
        str
            String of the ProbabilityDistribution object.
        """
        return f"N: {self._n}\n" \
               f"Init State:\n\t {self._initState}\n" \
               f"Operator:\n\t{self._operator}\n"\
               f"Final State:\n\t{self._finalState}"



from __future__ import annotations

import numpy as np
from qutip import Qobj, Options, mesolve
from qwak.StochasticOperator import StochasticOperator
from qwak.State import State


class StochasticQuantumWalk(object):
    """_summary_"""

    def __init__(
            self,
            state: State,
            operator: StochasticOperator) -> None:
        """This object is initialized with a user inputted initial state and
        operator.
        The dimension of the quantum walk will then be loaded from the initial
        state.
        The final state will contain the amplitudes of the time evolution of
        the initial state, as a function of the operator. This variable is initialized
        as an instance of QObj class.

        Parameters
        ----------
        state : State
            Initial state which will be the basis of the time dependant evolution.
        operator : StochasticOperator
            Operator which will evolve the initial state.
        """
        self._n = state.getDim()
        self._initState = state
        self._initQutipState = Qobj(state.getStateVec())
        self._operator = operator
        self._finalState = Qobj(State(self._n))
        self._time = 0

    def buildWalk(
        self,
        time,
        observables=[],
        opts=Options(store_states=True, store_final_state=True),
    ) -> None:
        """_summary_

        Parameters
        ----------
        time : _type_
            _description_
        observables : list, optional
            _description_, by default []
        opts : _type_, optional
            _description_, by default Options(store_states=True, store_final_state=True)
        """
        # TODO: Can we move the time dependency to the StochasticOperator class?
        # TODO: Can we make the time evolution low cost?
        # TODO: Is there a way to obtain amplitudes?
        # TODO: The final state is a of the Qobj class. Find a way to
        # make it State class.
        self._time = np.arange(0, time + 1)
        if self._operator.getSinkNode() is not None:
            self._initQutipState = Qobj(
                np.vstack([self._initState.getStateVec(), [0.0]])
            )
        self._finalState = mesolve(
            self._operator.getQuantumHamiltonian(),
            self._initQutipState,
            self._time,
            self._operator.getClassicalHamiltonian(),
            observables,
            options=opts,
        ).final_state.full()
        # if you want the full list of states, keep option store_states=True and instead
        # of final_state.full() use states.full()

    def getFinalState(self) -> Qobj:
        """_summary_

        Returns
        -------
        Qobj
            _description_
        """
        return self._finalState

    def setFinalState(self, newFinalState) -> None:
        """_summary_

        Parameters
        ----------
        newFinalState : _type_
            _description_
        """
        self._finalState = newFinalState

    def getDim(self) -> int:
        """_summary_

        Returns
        -------
        int
            _description_
        """
        return self._n
